/////////////////////////////////////////////////////////////////////////////
//  Resource File for : SpaceWar
#include "SpaceWarRes.h"

// The following line is a workaround for a redefinition warning in sal.h
// when using Visual Studio 2005 compilers and includes.  Comment out the
// line if you would like to see the warning.
#define _INC_CRTDEFS


/////////////////////////////////////////////////////////////////////////////
// This resource files requires the Platform SDK to be compiled.
// #include <gameux.h>

// These are the only two definitions needed from gameux.h
#define ID_GDF_XML __GDF_XML
#define ID_GDF_THUMBNAIL __GDF_THUMBNAIL

#define APSTUDIO_READONLY_SYMBOLS

#include <windows.h>

#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Neutral Resources
#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_NEU)
#ifdef _WIN32

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
#pragma code_page(DEFAULT)
#endif //_WIN32

ID_GDF_XML           DATA     "NEU\\SpaceWar.gdf.xml"
ID_GDF_THUMBNAIL     DATA     "NEU\\boxart_NEU.png"
ID_ICON              ICON     "NEU\\gameicon_NEU.ico"

#endif    // Neutral resources
/////////////////////////////////////////////////////////////////////////////
