//========= Copyright © 1996-2008 Valve LLC, All rights reserved. ============
//
// Purpose: 
//
// $NoKeywords: $
//=============================================================================

// stdafx.cpp : source file that includes just the standard includes
// SteamworksExample.pch will be the pre-compiled header
// stdafx.obj will contain the pre-compiled type information

#include "stdafx.h"
#include <stdio.h>
#ifdef WIN32
#include <varargs.h>
#include <tchar.h>
#endif

// TODO: reference any additional headers you need in STDAFX.H
// and not in this file
